/* File:      cov_mat.C
   Date:      5/19/03 (originally written in fall 2001)
   Author:    C. Paciorek
   Purpose:   code for creating, factoring and multiplying positive definite matrices
   Compile:   g++ -c -Wall -ansi
   Usage:     N/A
   Arguments: N/A
   Requires:  see just below
   Reference: data structures and code for matrices and manipulations follow George & Liu 1981 as I originally was using sparse calculations, which I'm not doing now
*/

/*
@Book{Geor:Liu:1981,
Author={George, A. and J.W.-H. Liu},
Title={Computer solution of large sparse positive definite systems},
Year=1981,
Address={Englewood Cliffs, New Jersey},
Publisher={Prentice-Hall, Inc.},
Pages=324
}
*/

#include "base.h"
#include "cov_mat.h"
#include "covariates.h"
#include "corr.h"

/////////////////////////////
// Constructors
//////////////////////////////

cov_mat_calc::cov_mat_calc(int size_in){
  size=size_in;
  sym_chol();              
}

sparse_cov_mat::sparse_cov_mat(int size_in,int num_covars_in,double* smoothness_in){
  size=size_in;
  num_covars=num_covars_in;
  smoothness=smoothness_in;
}


sparse_cov_mat_sum::sparse_cov_mat_sum(int size_in,double* fun_lnz_R_in,double* fun_lnz_R_stored_in,double* fun_sigma_in,double* fun_sigma_stored_in,double* data_sigma_in,double* fun_in,double* num_trials_in):sparse_cov_mat(size_in,0,0){
  num_trials=num_trials_in;
  fun=fun_in;
  fun_sigma=fun_sigma_in;
  fun_sigma_stored=fun_sigma_stored_in;
  data_sigma=data_sigma_in;
  fun_lnz_R=fun_lnz_R_in;
  fun_lnz_R_stored=fun_lnz_R_stored_in;
  sym_chol();
  Cfmat=new mat_pp<double>(size,size);
}

sparse_cov_mat_stat::sparse_cov_mat_stat(int size_in,int num_covars_in,double** kappa_in,double** gamma_in,double* smoothness_in) : sparse_cov_mat(size_in,num_covars_in,smoothness_in){
  kappa=kappa_in;  
  gamma=gamma_in;  
  if(num_covars>1){
    kernel=new double[num_covars*num_covars];
    kernel_mat=new cov_mat_calc(num_covars);
    vec = new double[num_covars];
  }
  sym_chol();                     // create Cholesky data structure
  fill(FIRST);                    // fill matrix 
  gen_chol(NO_WARN);              // do numerical factorization
  update(ACCEPT);                 // fill stored vectors 
}


sparse_cov_mat_nonstat::sparse_cov_mat_nonstat(int size_in,int num_covars_in,double** eigenval_in,double** eigenvec_in,double* smoothness_in): sparse_cov_mat(size_in,num_covars_in,smoothness_in){

  eigenval=eigenval_in;
  eigenvec=eigenvec_in;

  // eigenvector matrices for all covariates
  eigenvecs=new mat_pp<double>(size,num_covars*num_covars);
  eigenvecs_stored=new mat_pp<double>(size,num_covars*num_covars);

  // individual kernel matrix
  if(num_covars>1){
    kernel=new mat_pp<double>(size,num_covars*num_covars);
    kernel_sum=new double[num_covars*num_covars];
    vec=new double[num_covars];
    kernel_sum_mat=new cov_mat_calc(num_covars);
  }
  sym_chol();                     // create Cholesky data structure
  fill(FUN_COVAR_EIGVEC,FIRST);   // fill matrix
  gen_chol(NO_WARN);              // do numerical factorization
  update(ACCEPT);                 // fill stored vectors
}

/////////////////////////////////////
// Destructors 
////////////////////////////////////

cov_mat_calc::~cov_mat_calc(void){

  delete [] nzsub;
  delete [] xlnz;
  delete [] xnzsub;
  delete [] lnz_L;
  delete [] lnz_C;
  delete [] diag_L;
  delete [] diag_C;
  delete [] lnz_L_stored; 
  delete [] diag_L_stored;
  delete [] lnz_C_stored; 
  delete [] diag_C_stored;
}

sparse_cov_mat::~sparse_cov_mat(void){

  delete [] nzsub;
  delete [] xlnz;
  delete [] xnzsub;
  delete [] lnz_L;
  delete [] lnz_R;
  delete [] diag_L;
  delete [] diag_R;
  delete [] lnz_L_stored; 
  delete [] diag_L_stored;
  delete [] lnz_R_stored; 
  delete [] diag_R_stored;
  delete euc_distance;
}

sparse_cov_mat_sum::~sparse_cov_mat_sum(void){
  delete Cfmat;
}

sparse_cov_mat_stat::~sparse_cov_mat_stat(void){
  if(num_covars>1){
    delete [] kernel;
    delete kernel_mat;
    delete [] vec;
  }
}

sparse_cov_mat_nonstat::~sparse_cov_mat_nonstat(void)
{ 
  delete eigenvecs;
  delete eigenvecs_stored;
  if(num_covars>1){
    delete kernel;
    delete [] kernel_sum;
    delete [] vec;
    delete kernel_sum_mat;
  }
}

///////////////////////////////////////////
// Filling routines
//////////////////////////////////////////

void sparse_cov_mat_sum::fill(void){
  // fill sum of covariance matrices (despite it being named "R") from data covariance and prior function covariance
  int index;
  for(index=0;index<size;index++){
    diag_R[index]=(*fun_sigma)*(*fun_sigma)*(1.0)+(1.0/(num_trials[index]*fun[index]*(1-fun[index])));
  }
  for(index=0;index<maxlnz;index++){
    lnz_R[index]=fun_lnz_R[index]*(*fun_sigma)*(*fun_sigma);
  }
}

void sparse_cov_mat_sum::fill_old(void){
  int index;
  for(index=0;index<size;index++){
    //      diag_R[i]=fillij(locations->perm[i]-1,locations->perm[i]-1);
    diag_R[index]=(*fun_sigma_stored)*(*fun_sigma_stored)*(1.0)+(1.0/(num_trials[index]*fun[index]*(1-fun[index])));
  }
  // fill off-diags in sparse structure -- see details above
  for(index=0;index<maxlnz;index++){
    lnz_R[index]=fun_lnz_R_stored[index]*(*fun_sigma_stored)*(*fun_sigma_stored);
  }
}

void cov_mat_calc::fill(double* input){
  // fill matrix from vector representation
  int col,istart,istop,i,isub,index,ii;
  index=0;
  for(col=0;col<size;col++){
    diag_C[col]=input[col*size+col];
  }
  for(col=0;col<size-1;col++){  
    istart=xlnz[col];  
    istop=xlnz[col+1]-1;
    i=xnzsub[col];
    for(ii=istart;ii<=istop;ii++){
      isub=nzsub[i-1]-1;
      lnz_C[index]=input[isub*size+col];
      i++; index++;
    }
  }
}

void sparse_cov_mat_stat::fill(int first)
{
  int index, col, i, istart, istop, ii, isub;

  // correlation matrix has ones on diagonal
  if(first==FIRST){
    for(index=0;index<size;index++){
      diag_R[index]=1.0;
    }
  }

  // only implemented for 1-3 covariates
  // construct kernel matrix used for all covariates (i.e., Mahalanobis distance matrix
  if(num_covars==2){
    // kappa values are square roots of eigenvalues
    double lam1=(*(kappa[0]))*(*(kappa[0]));
    double lam2=(*(kappa[1]))*(*(kappa[1]));
    // Givens angle
    double cos1=cos(*(gamma[0]));
    double sin1=sin(*(gamma[0]));
    // calculate kernel matrix
    kernel[0]=lam1*cos1*cos1+lam2*sin1*sin1;
    kernel[1]=(lam1-lam2)*cos1*sin1;
    kernel[2]=kernel[1];
    kernel[3]=lam1*sin1*sin1+lam2*cos1*cos1;
  }
  if(num_covars==3){
    // Givens angles
    double c12=cos(*(gamma[0]));
    double s12=sin(*(gamma[0]));
    double c13=cos(*(gamma[1]));
    double s13=sin(*(gamma[1]));
    double c23=cos(*(gamma[2]));
    double s23=sin(*(gamma[2]));
    double a,b,c,d,e,f,g,h,i;
    a=c12*c13;
    b=-s12*c23-c12*s13*s23;
    c=s12*s23-c12*s13*c23;
    d=s12*c13;
    e=c12*c23-s12*s13*s23;
    f=-c12*s23-s12*s13*c23;
    g=s13;
    h=c13*s23;
    i=c13*c23;
    // eigenvalues
    double lam1=(*(kappa[0]))*(*(kappa[0]));
    double lam2=(*(kappa[1]))*(*(kappa[1]));
    double lam3=(*(kappa[2]))*(*(kappa[2]));
    // calculate kernel matrix
    kernel[0]=a*a*lam1+b*b*lam2+c*c*lam3;
    kernel[1]=a*d*lam1+b*e*lam2+c*f*lam3;
    kernel[3]=kernel[1];
    kernel[2]=a*g*lam1+b*h*lam2+c*i*lam3;
    kernel[6]=kernel[2];
    kernel[4]=d*d*lam1+e*e*lam2+f*f*lam3;
    kernel[5]=d*g*lam1+e*h*lam2+f*i*lam3;
    kernel[7]=kernel[5];
    kernel[8]=g*g*lam1+h*h*lam2+i*i*lam3;
    kernel_mat->fill(kernel);  // create matrix representation
    kernel_mat->gen_chol(WARN);  // find Cholesky decomposition
  }
  index=0;
  
  if(first==FIRST){
    // calculate Euclidean distance between all covariates
    for(col=0;col<size-1;col++){  
      istart=xlnz[col];  
      istop=xlnz[col+1]-1;
      i=xnzsub[col];
      for(ii=istart;ii<=istop;ii++){
        isub=nzsub[i-1]-1;
        covariates->calc_datadata_distance(isub,col,euc_distance->vals[index],PERMUTE);  // Note that euc_distance is initialized in sym_chol()
        lnz_R[index]=fillij(index);  // fill individual elements of matrix
        i++; index++;
      }
    }
  }
  else{
    for(col=0;col<size-1;col++){  
      istart=xlnz[col];  
      istop=xlnz[col+1]-1;
      i=xnzsub[col];
      for(ii=istart;ii<=istop;ii++){
        isub=nzsub[i-1]-1;
        lnz_R[index]=fillij(index);  // fill individual elements
        i++; index++;
      }
    }
  }
}

double sparse_cov_mat_stat::fillij(int index){
  // fill individual elements of stationary correlation matrix

  // calculate quadratic form
  double Q=0.0;
  if(num_covars==1){
    Q=pow(euc_distance->vals[index][0]/(*(kappa[0])),2.0);
  }
  if(num_covars==2){ // dist^T * Kernel matrix * dist
    double det=kernel[0]*kernel[3]-kernel[1]*kernel[1];
    Q=(kernel[3]*euc_distance->vals[index][0]*euc_distance->vals[index][0]-2.0*kernel[1]*euc_distance->vals[index][0]*euc_distance->vals[index][1]+kernel[0]*euc_distance->vals[index][1]*euc_distance->vals[index][1])/det;
  }
  if(num_covars==3){  // dist^T * Kernel matrix * dist
    kernel_mat->Csolve(euc_distance->vals[index],vec); 
    for(int p=0;p<num_covars;p++){
      Q+=vec[p]*euc_distance->vals[index][p];
    }
  }

  // calculate correlation; kappa(scale parameter) is 1 since it's folded into Q
  return(matern_corr_euc(sqrt(Q),1.0,*smoothness));
}

void sparse_cov_mat_nonstat::fill(int which,int first){
  // fills individual values of nonstationary correlation matrix

  int j, index,istart,istop,ii,col,isub,p;
  double len;
  /*  the following code walks through the sparse structure of the matrix by row within column
      istart and istop determine how many non-zero values are in each column
      index walks through all the non-zero values one by one
      xnzsub tells where in nzsub the row values for the current column start
      nzsub contains the row values for every non-zero element
      values within xnsub and nzsub are 1,..., hence the -1's sprinkled 
      throughout to get C style 0,... indexing
      note that some 'non-zero' values might be zero
      I may want to consider changing isub to row, nzsub to row_vals and xnzsub to row_vals_pointer
  */

  // calculate kernel matrices
  if((which==FUN_COVAR_EIGVEC || which==KERNEL) && num_covars>1){
    for(j=0;j<size;j++){  
      // calculate eigenvector matrices from eigenvector coordinates - filled by row
      if(num_covars==2){
        len=pow(eigenvec[0][j]*eigenvec[0][j]+eigenvec[1][j]*eigenvec[1][j],0.5);
        eigenvecs->vals[j][0]=eigenvec[0][j]/len;
        eigenvecs->vals[j][2]=eigenvec[1][j]/len;
        eigenvecs->vals[j][1]=-eigenvecs->vals[j][2];
        eigenvecs->vals[j][3]=eigenvecs->vals[j][0];
      }
      if(num_covars==3){
        double x1,x2,x3,y1,y2,y3,z1,z3,u1,v1,u2,v2,lxyz,lxy,luv;
        lxyz=pow(eigenvec[0][j]*eigenvec[0][j]+eigenvec[1][j]*eigenvec[1][j]+eigenvec[2][j]*eigenvec[2][j],0.5);
        x1=eigenvec[0][j]/lxyz;
        y1=eigenvec[1][j]/lxyz;
        z1=eigenvec[2][j]/lxyz;
        luv=pow(eigenvec[3][j]*eigenvec[3][j]+eigenvec[4][j]*eigenvec[4][j],0.5);
        u1=eigenvec[3][j]/luv;
        v1=eigenvec[4][j]/luv;
        u2=-v1;
        v2=u1;
        // eigenvector matrix is product of initial matrix defining first eigenvector (x1,y1,z1) and a rotation in the p-1 dimensional space orthogonal to the first eigenvector
        lxy=pow(eigenvec[0][j]*eigenvec[0][j]+eigenvec[1][j]*eigenvec[1][j],0.5);
        y2=eigenvec[0][j]/lxy;
        x2=-eigenvec[1][j]/lxy;
        if(sqrt(x1*x1+y1*y1)<0.00001){
          y2=1.0;
          x2=0.0;
          cerr << "Warning: eigenvector may not have been proposed smoothly\n";
          cerr.flush();
        }
        x3=-eigenvec[0][j]*eigenvec[2][j]/(lxy*lxyz);
        y3=-eigenvec[1][j]*eigenvec[2][j]/(lxy*lxyz);
        z3=lxy/lxyz;
        eigenvecs->vals[j][0]=x1;
        eigenvecs->vals[j][3]=y1;
        eigenvecs->vals[j][6]=z1;
        eigenvecs->vals[j][1]=u1*x2+v1*x3;
        eigenvecs->vals[j][2]=u2*x2+v2*x3;
        eigenvecs->vals[j][4]=u1*y2+v1*y3;
        eigenvecs->vals[j][5]=u2*y2+v2*y3;
        eigenvecs->vals[j][7]=v1*z3;
        eigenvecs->vals[j][8]=v2*z3;

        // check for problems
        double len1,len2,len3;
        len1=x1*x1+y1*y1+z1*z1; len2=x2*x2+y2*y2; len3=x3*x3+y3*y3+z3*z3;
        // are eigenvectors unit length?
        if(fabs(len1-1.0)+fabs(len2-1.0)+fabs(len3-1.0)>.0001){
          cout << "Non-unit length eigenvectors: " << x1 << " " << y1 << " " << z1 << " " << x2 << " " << y2 << " " << x3 << " " << y3 << " " << z3 << endl;
          cout.flush();
        }
        // are eigenvectors orthogonal?
        len1=x1*x2+y1*y2; len2=x1*x3+y1*y3+z1*z3; len3=x2*x3+y2*y3;
        if(fabs(len1)+fabs(len2)+fabs(len3)>0.0001){
          cout << "Non-orthogonal eigenvectors: " << x1 << " " << y1 << " " << z1 << " " << x2 << " " << y2 << " " << x3 << " " << y3 << " " << z3 << endl;
          cout.flush();
        }
      }
    }
  }

  if(num_covars>1){
    for(j=0;j<size;j++){ 
      // direct calculation of kernel matrices as Gamma*Lambda*t(Gamma) from eigendecomposition
      if(num_covars==2){
        double a,b,c,d;
        a=eigenvecs->vals[j][0];
        b=eigenvecs->vals[j][1];
        c=eigenvecs->vals[j][2];
        d=eigenvecs->vals[j][3];
        kernel->vals[j][0]=a*a*eigenval[0][j]+b*b*eigenval[1][j];
        kernel->vals[j][1]=a*c*eigenval[0][j]+b*d*eigenval[1][j];
        kernel->vals[j][2]=kernel->vals[j][1];
        kernel->vals[j][3]=c*c*eigenval[0][j]+d*d*eigenval[1][j];
      }
      if(num_covars==3){
        double a,b,c,d,e,f,g,h,i;
        a=eigenvecs->vals[j][0];
        b=eigenvecs->vals[j][1];
        c=eigenvecs->vals[j][2];
        d=eigenvecs->vals[j][3];
        e=eigenvecs->vals[j][4];
        f=eigenvecs->vals[j][5];
        g=eigenvecs->vals[j][6];
        h=eigenvecs->vals[j][7];
        i=eigenvecs->vals[j][8];
        kernel->vals[j][0]=a*a*eigenval[0][j]+b*b*eigenval[1][j]+c*c*eigenval[2][j];
        kernel->vals[j][1]=a*d*eigenval[0][j]+b*e*eigenval[1][j]+c*f*eigenval[2][j];
        kernel->vals[j][3]=kernel->vals[j][1];
        kernel->vals[j][2]=a*g*eigenval[0][j]+b*h*eigenval[1][j]+c*i*eigenval[2][j];
        kernel->vals[j][6]=kernel->vals[j][2];
        kernel->vals[j][4]=d*d*eigenval[0][j]+e*e*eigenval[1][j]+f*f*eigenval[2][j];
        kernel->vals[j][5]=d*g*eigenval[0][j]+e*h*eigenval[1][j]+f*i*eigenval[2][j];
        kernel->vals[j][7]=kernel->vals[j][5];
        kernel->vals[j][8]=g*g*eigenval[0][j]+h*h*eigenval[1][j]+i*i*eigenval[2][j];
      }
      
      // check kernel matrices
      for(p=0;p<num_covars;p++){
        if(kernel->vals[j][p*num_covars+p]<=0.0){
          cout << "Warning: negative variances\n"; cout.flush();
        }
      }
      if(num_covars==2){
        if(fabs(kernel->vals[j][1])>=sqrt(kernel->vals[j][0]*kernel->vals[j][3])){
          cout << "Warning: kernel matrices not positive definite " << j << " " << kernel->vals[j][0] << " " << kernel->vals[j][1] << " " << kernel->vals[j][3] << " " << eigenvecs->vals[j][0] << " " << eigenvecs->vals[j][1] << " " << eigenvecs->vals[j][2] << " " << eigenvecs->vals[j][3] << " " << eigenval[0][j] << " " << eigenval[1][j] << endl; cout.flush();
        }
      }
    }
  }

  // fill correlation matrix
  if(first==FIRST){
    for(index=0;index<size;index++){
      // correlation matrix has ones on diagonal
      diag_R[index]=1.0;                        
    }
    // fill off-diags in sparse structure -- see details above
    index=0;
    for(col=0;col<size-1;col++){  
      istart=xlnz[col];  
      istop=xlnz[col+1]-1;
      j=xnzsub[col];
      for(ii=istart;ii<=istop;ii++){
	isub=nzsub[j-1]-1;
	// calculate and store Euclidean distances between all pairs of covariates
	covariates->calc_datadata_distance(isub,col,euc_distance->vals[index],PERMUTE);
	lnz_R[index]=fillij(isub,col,index);  // find correlation matrix element
	j++; index++;
      }
    }
  }
  else{  
    index=0;
    for(col=0;col<size-1;col++){  
      istart=xlnz[col];  
      istop=xlnz[col+1]-1;
      j=xnzsub[col];
      for(ii=istart;ii<=istop;ii++){
	isub=nzsub[j-1]-1;
	lnz_R[index]=fillij(isub,col,index);  // find correlation matrix element
	j++; index++;
      }
    }
  }
}


double sparse_cov_mat_nonstat::fillij(int i, int j,int index){
  double Q=0.0;
  // log determinant values
  double deti=0.0; double detj=0.0; double detsum=0.0;
  int p;
  double outval=0.0;

  if(num_covars>1){
    for(p=0;p<num_covars;p++){
      deti+=log(eigenval[p][i]);
      detj+=log(eigenval[p][j]);
    }
    // calculate sum of kernels
    for(p=0;p<num_covars*num_covars;p++){
      kernel_sum[p]=kernel->vals[i][p]+kernel->vals[j][p];
    }
    
    // direct calculation
    if(num_covars==2){
      double det=kernel_sum[0]*kernel_sum[3]-kernel_sum[1]*kernel_sum[1];
      detsum=log(sqrt(det));      
      // calculate quadratic form
      Q=(kernel_sum[3]*euc_distance->vals[index][0]*euc_distance->vals[index][0]-2.0*kernel_sum[1]*euc_distance->vals[index][0]*euc_distance->vals[index][1]+kernel_sum[0]*euc_distance->vals[index][1]*euc_distance->vals[index][1])/det;
    }
    if(num_covars==3){
      // use cov_mat_calc to do calculation of quadratic form and determinant
      kernel_sum_mat->fill(kernel_sum);  // create matrix representation
      kernel_sum_mat->gen_chol(WARN);    
      kernel_sum_mat->Csolve(euc_distance->vals[index],vec); 
      for(p=0;p<num_covars;p++){
        Q+=vec[p]*euc_distance->vals[index][p];
      }
      detsum=kernel_sum_mat->det_log_squ_root();
    }
    // calculate nonstationary Matern
    outval=(num_covars/2.0)*log(2.0)+0.25*deti+0.25*detj-detsum+log(matern_corr_euc(sqrt(2.0*Q),1.0,*smoothness));  // 2.0*Q causes averaging rather than summing of kernels
     return(exp(outval));
  }
  else{  // one-dimensional covariate
    double sig2i=eigenval[0][i];  // kernel matrix is just the variance
    double sig2j=eigenval[0][j];  // ... same 
    // calculate nonstationary Matern
    return(sqrt(2.0)*pow(sig2i*sig2j,0.25)*matern_corr_euc(fabs(euc_distance->vals[index][0])/sqrt((sig2i+sig2j)/2.0),1.0,*smoothness)/(sqrt(sig2i+sig2j)));
  }
}


double sparse_cov_mat_sum::calc_df(void){
  // df for smoothing matrix
  int col,index,istart,istop,isub,i,ii;
  double* tmpvec=new double[size];
  double trace=0.0;

  // create vector of columns from C_f prior covariance matrix -- need to do this because can't use Csolve applied to the lnz/diag matrix representation
  for(col=0;col<size;col++){  
    Cfmat->vals[col][col]=1.0;
  }
  index=0;
  for(col=0;col<size-1;col++){  
    istart=xlnz[col];  
    istop=xlnz[col+1]-1;
    i=xnzsub[col];
    for(ii=istart;ii<=istop;ii++){
      isub=nzsub[i-1]-1;
      Cfmat->vals[isub][col]=fun_lnz_R[index];
      Cfmat->vals[col][isub]=fun_lnz_R[index];
      i++; index++;
    }
  }

  for(col=0;col<size;col++){
    Csolve(Cfmat->vals[col],tmpvec);  // do (C_f+C_y)^{-1} applied to each column of C_f
    trace+=tmpvec[col];
  }
  trace*=(*fun_sigma)*(*fun_sigma);  // include function variance (sigma^2)
  trace++;       // add one for effect of estimating mu for function
  delete [] tmpvec;
  return(trace);
}

void sparse_cov_mat::update(int accepted){
  // if accepted, make stored values the newly-accepted values, o.w. make stored values the current values
  if(accepted){
    copy(size,diag_L,diag_L_stored);
    copy(maxlnz,lnz_L,lnz_L_stored);
    copy(maxlnz,lnz_R,lnz_R_stored);
    copy(size,diag_R,diag_R_stored);
  }
  else{
    copy(size,diag_L_stored,diag_L);
    copy(maxlnz,lnz_L_stored,lnz_L);
    copy(maxlnz,lnz_R_stored,lnz_R);
    copy(size,diag_R_stored,diag_R);
  }
}

void sparse_cov_mat_nonstat::update(int accepted){
  // if accepted, make stored values the newly-accepted values, o.w. make stored values the current values
  int i;
  sparse_cov_mat::update(accepted);

  // save eigenvecs to be used when only eigenvalues are proposed
  if(accepted){
    if(num_covars>1){
      for(i=0;i<size;i++){    
	copy(num_covars*num_covars,eigenvecs->vals[i],eigenvecs_stored->vals[i]);
      }
    }
  }
  else{
    if(num_covars>1){
      for(i=0;i<size;i++){
	copy(num_covars*num_covars,eigenvecs_stored->vals[i],eigenvecs->vals[i]);
      }
    }
  }
}


void sparse_cov_mat::sym_chol(void)
{
  // calculate symbolic Cholesky factorization -- i.e., set up sparse structure based on adjacency information
  int* rchlnk; int* mrglnk; int* marker; int* temp_nzsub; 

  rchlnk = new int[size];  // temp storage for use by Fortran routine
  mrglnk=new int[size];    //  "
  marker=new int[size];    //  "
  temp_nzsub=new int[size*size];   // row indices
  xlnz=new int[size+1];            // pointers to values (where each column starts)
  xnzsub=new int[size+1];          // pointers to row indices (where each column starts)
  maxsub = size*size;  // input size of nzsub -- rewritten within smbfct_

  covariates->choose_adjinfo(size);  // select adjacency information for either just the training set or for the training plus test set
  smbfct_(&size,covariates->xadj,covariates->adjncy,covariates->perm,covariates->invp,xlnz,&maxlnz,xnzsub,temp_nzsub,&maxsub,rchlnk,mrglnk, marker);  // call to Fortran to set up the structure
 
  nzsub=new int[maxsub];
  copy(maxsub,temp_nzsub,nzsub);
  lnz_L=new double[maxlnz];
  lnz_R=new double[maxlnz];
  diag_L=new double[size];
  diag_R=new double[size];
  lnz_L_stored=new double[maxlnz];
  diag_L_stored=new double[size];
  lnz_R_stored=new double[maxlnz];
  diag_R_stored=new double[size];

  euc_distance=new mat_pp<double>(maxlnz,num_covars);
  // initialize structure to hold pairwise distances between covariates

  delete [] rchlnk;
  delete [] mrglnk;
  delete [] marker;
  delete [] temp_nzsub;
}

void cov_mat_calc::sym_chol(void){
    // calculate symbolic Cholesky factorization -- i.e., set up sparse structure based on adjacency information
  int* rchlnk; int* mrglnk; int* marker; int* temp_nzsub; 

  rchlnk = new int[size];  // temp storage for use by Fortran routine
  mrglnk=new int[size];    //  "
  marker=new int[size];    //  "
  temp_nzsub=new int[size*size];   // row indices
  xlnz=new int[size+1];            // pointers to values (where each column starts)
  xnzsub=new int[size+1];          // pointers to row indices (where each column starts)
  maxsub = size*size;  // input size of nzsub -- rewritten within smbfct_

  // create adjacency information for use by smbfct_
  xadj=new int[size+1];
  adjncy=new int[size*size+1];
  int cnt=0;
  for(int p=0;p<size;p++){
    for(int q=0;q<size;q++){
      adjncy[cnt]=q+1;
      if(q==0){
        xadj[p]=cnt+1;
      }
      cnt++;
    }
  }
  adjncy[cnt]=0;
  xadj[size]=size*size+1;

  int* tmpvec=new int[size];
  for(int j=0;j<size;j++){
    tmpvec[j]=j+1;
  }

  smbfct_(&size,xadj,adjncy,tmpvec,tmpvec,xlnz,&maxlnz,xnzsub,temp_nzsub,&maxsub,rchlnk,mrglnk, marker);  // call Fortran to set up structure
 
  nzsub=new int[maxsub];
  copy(maxsub,temp_nzsub,nzsub);
  lnz_L=new double[maxlnz];
  lnz_C=new double[maxlnz];
  diag_L=new double[size];
  diag_C=new double[size];
  lnz_L_stored=new double[maxlnz];
  diag_L_stored=new double[size];
  lnz_C_stored=new double[maxlnz];
  diag_C_stored=new double[size];

  delete [] adjncy;
  delete [] xadj;
  delete [] tmpvec;
  delete [] rchlnk;
  delete [] mrglnk;
  delete [] marker;
  delete [] temp_nzsub;
}

void sparse_cov_mat::gen_chol(int warn){
  // calculate numerical Cholesky factorization; if warn==WARN then the Fortran routine says when a column is zeroed out

  double eps=TOL;  // this is the value of the squared diagonal below which a column is zeroed out
  int *first; int* link; double* temp;
  first=new int[size];
  link=new int[size];
  temp=new double[size];

  // temporary vectors are necessary because gsfct_ overwrites input vectors
  double* temp_lnz_L=new double[maxlnz];
  double* temp_diag_L=new double[size];
  copy(maxlnz,lnz_R,temp_lnz_L);
  copy(size,diag_R,temp_diag_L);

  // Fortran routine for calculating generalized Cholesky
  gsfct_(&size,xlnz,temp_lnz_L,xnzsub,nzsub,temp_diag_L,link,first,temp,&eps,&warn);  

  copy(maxlnz,temp_lnz_L,lnz_L);
  copy(size,temp_diag_L,diag_L);

  delete [] first;
  delete [] link;
  delete [] temp;
  delete [] temp_lnz_L;
  delete [] temp_diag_L;
}


void cov_mat_calc::gen_chol(int warn){
  // calculate numerical Cholesky factorization; if warn==WARN then the Fortran routine says when a column is zeroed out
  double eps=TOL;
  int *first; int* link; double* temp;
  first=new int[size];
  link=new int[size];
  temp=new double[size];

  // temporary vectors are necessary because gsfct_ overwrites input vectors
  // and because gsfct_ uses floats not doubles - but it could use double no?  why do I use floats? faster?

  double* temp_lnz_L=new double[maxlnz];
  double* temp_diag_L=new double[size];
  copy(maxlnz,lnz_C,temp_lnz_L);
  copy(size,diag_C,temp_diag_L);

  // Fortran routine for calculating generalized Cholesky
  gsfct_(&size,xlnz,temp_lnz_L,xnzsub,nzsub,temp_diag_L,link,first,temp,&eps,&warn);

  copy(maxlnz,temp_lnz_L,lnz_L);
  copy(size,temp_diag_L,diag_L);

  delete [] first;
  delete [] link;
  delete [] temp;
  delete [] temp_lnz_L;
  delete [] temp_diag_L;
}

double cov_mat_calc::det_log_squ_root(void){
  // calculates log square root of determinant as log of product of diagonals of Cholesky factor
  int i;
  double tmp=0.0;
  for(i=0;i<size;i++){
    if(diag_L[i]>=TOL){
      tmp+=log(diag_L[i]);
    }
    else{  // kludge for case when diagonal is zeroed out and determinant is zero
      tmp+=log(TOL);
      cerr << "Warning: used tolerance in place of true diagonal value in calculating determinant\n";
    }
  }
  return(tmp);
}

double sparse_cov_mat::det_log_squ_root(void){
  // calculates log square root of determinant as log of product of diagonals of Cholesky factor
  int i;
  double tmp=0.0;
  for(i=0;i<size;i++){
    if(diag_L[i]>=TOL){
      tmp+=log(diag_L[i]);
    }
    else{   // kludge for case when diagonal is zeroed out and determinant is zero
      tmp+=log(TOL);
      cerr << "Warning: used tolerance in place of true diagonal value in calculating determinant\n";
    }
  }
  return(tmp);
}

int sparse_cov_mat::Csolve(double* b, double* output){
  // calculates C^{-1} dot b
  // this follows the code of George & Liu's gsslv, p. 156-7

  // to solve Cb=L*t(L)b=y, let z=t(L)b, solve Lz=y for z; then solve
  // t(L)b=z for b;  i.e. y=C^{-1}b

  // when a column is zeroed out, keep output_i=old_output_i

  double rhsj, s;
  int i, ii, istop, istart,isub,j;

  double* rhs=new double[size];
  double* diag=new double[size]; 
  double* lnz=new double[maxlnz];

  copy(maxlnz,lnz_L,lnz);  
  copy(size,diag_L,diag);  
  copy(size,b,rhs);
  
  // forward substitution -- solve Lz=y

  int flag=0;
  for(j=0;j<size;j++){
    if(diag[j]<TOL){
      rhs[j]=0.0;    // for Lsolve, I have a candidate value of omega from previous value but here, there is no obvious candidate, so use 0
      if(flag==0){  // note that a column is zeroed out
	flag=j;
      }
    }
    else{
      rhsj= rhs[j]/diag[j];
      rhs[j]= rhsj;
      if(j==size-1 ) {break;}
      istart=xlnz[j];  
      if(j==size-2) {istop=istart;}
      else{
	istop=xlnz[j+1]-1;
      }
      i=xnzsub[j];
      for(ii=istart;ii<=istop;ii++){
	isub=nzsub[i-1];
	rhs[isub-1]=rhs[isub-1]-lnz[ii-1]*rhsj;
	i++;
      }
    }
  }
  if(flag){
        cout << "Warning: zeroed out column in C matrix chol: " << flag << "\n"; cout.flush();
  }

  // backward substitution -- solve t(L)b=z
  for(j=size-1;j>=0;j--) {
    if(diag[j]<TOL){
      rhs[j]=0.0;
    }
    else{
      s=rhs[j];
      if(j<size-1)
	{
	  istart=xlnz[j];
	  if(j==size-2) {istop=istart;}
	  else
	    {
	      istop=xlnz[j+1]-1;
	    }
	  i=xnzsub[j];
	  for(ii=istart;ii<=istop;ii++)
	    {
	      isub=nzsub[i-1];
              s=s-lnz[ii-1]*rhs[isub-1];
	      i++;
	    }
	}
         rhs[j]=s/diag[j];
    }
  }
  
  copy(size,rhs,output);

  delete [] rhs;
  delete [] diag;
  delete [] lnz;
 
  return(flag);
}

int cov_mat_calc::Csolve(double* b, double* output){
  // calculates C^{-1} dot b 
  // this follows the code of George & Liu's gsslv, p. 156-7

  // to solve Cb=L*t(L)b=y, let z=t(L)b, solve Lz=y for z; then solve
  // t(L)b=z for b;  i.e. y=C^{-1}b

  double rhsj, s;
  int i, ii, istop, istart,isub,j;

  double* rhs=new double[size];
  double* diag=new double[size]; 
  double* lnz=new double[maxlnz];

  copy(maxlnz,lnz_L,lnz);  
  copy(size,diag_L,diag);  
  copy(size,b,rhs);
  
  // forward substitution -- solve Lz=y

  int flag=0;
  for(j=0;j<size;j++){
    if(diag[j]<TOL){
      rhs[j]=0.0;    // for Lsolve, I have a candidate value of omega from previous value but here, there is no obvious candidate, so use 0
      if(flag==0){
	flag=j;
      }
    }
    else{
      rhsj= rhs[j]/diag[j];
      rhs[j]= rhsj;
      if(j==size-1 ) {break;}
      istart=xlnz[j];  
      if(j==size-2) {istop=istart;}
      else{
	istop=xlnz[j+1]-1;
      }
      i=xnzsub[j];
      for(ii=istart;ii<=istop;ii++){
	isub=nzsub[i-1];
	rhs[isub-1]=rhs[isub-1]-lnz[ii-1]*rhsj;
	i++;
      }
    }
  }
  if(flag){
        cout << "Warning: zeroed out column in C matrix chol: " << flag << "\n"; cout.flush();
  }

  // backward substitution -- solve t(L)b=z
  for(j=size-1;j>=0;j--) {
    if(diag[j]<TOL){
      rhs[j]=0.0;
    }
    else{
      s=rhs[j];
      if(j<size-1)
	{
	  istart=xlnz[j];
	  if(j==size-2) {istop=istart;}
	  else
	    {
	      istop=xlnz[j+1]-1;
	    }
	  i=xnzsub[j];
	  for(ii=istart;ii<=istop;ii++)
	    {
	      isub=nzsub[i-1];
              s=s-lnz[ii-1]*rhs[isub-1];
              i++;
	    }
	}
      rhs[j]=s/diag[j];
    }
  }
  
  copy(size,rhs,output);

  delete [] rhs;
  delete [] diag;
  delete [] lnz;
 
  return(flag);
}

void sparse_cov_mat::Lcheck(double* newvec,double* oldvec){
  // replace elements of newvec whose column is zeroed out with elements of oldvec; this prevents updating of elements that have no influence on determining the process values
  for(int j=0;j<size;j++){
    if(diag_L[j]<TOL){
      newvec[j]=oldvec[j];
    }
  }
}


void sparse_cov_mat::Lmult(double* b,double* output)
{
  // calculates L*b
  int i, ii, istop, istart,j,row;
  for(j=0;j<size;j++){
      output[j]=0.0;
    }
  // j is column
  for(j=0;j<size-1;j++){
    if(diag_L[j]>=TOL){  // otherwise column is not used
      output[j]+=diag_L[j]*b[j];
      istart=xlnz[j];
      istop=xlnz[j+1]-1;
      i=xnzsub[j];
      for(ii=istart;ii<=istop;ii++){
	row=nzsub[i-1]-1;
	output[row]+=lnz_L[ii-1]*b[j];
	i++;
      }
    }
  }
  output[size-1]+=diag_L[size-1]*b[size-1];
}



void sparse_cov_mat::Ltmult(double* b,double* output){
  // calculates L^T*b

  int i, ii, istop, istart,j,row;
  for(j=0;j<size;j++){
      output[j]=0.0;
    }
  // j is column
  for(j=0;j<size-1;j++){
    if(diag_L[j]>=TOL){  // otherwise column is not used
      output[j]+=diag_L[j]*b[j];
      istart=xlnz[j];
      istop=xlnz[j+1]-1;
      i=xnzsub[j];
      for(ii=istart;ii<=istop;ii++){
	row=nzsub[i-1]-1;
	output[j]+=lnz_L[ii-1]*b[row];
	i++;
      }
    }
  }
  output[size-1]+=diag_L[size-1]*b[size-1];
}

void sparse_cov_mat::Ltmult_old(double* b,double* output){
  // calculates L^T*b for stored Cholesky

  int i, ii, istop, istart,j,row;
  for(j=0;j<size;j++){
      output[j]=0.0;
    }
  // j is col
  for(j=0;j<size-1;j++){
    if(diag_L_stored[j]>=TOL){
      output[j]+=diag_L_stored[j]*b[j];
      istart=xlnz[j];
      istop=xlnz[j+1]-1;
      i=xnzsub[j];
      for(ii=istart;ii<=istop;ii++){
	row=nzsub[i-1]-1;
	output[j]+=lnz_L_stored[ii-1]*b[row];
	i++;
      }
    }
  }
  output[size-1]+=diag_L_stored[size-1]*b[size-1];
}

void sparse_cov_mat::Rmult(double* b,double* output){
  // calculates R*b

  int i, ii, istop, istart,j,row;
  for(j=0;j<size;j++){
    output[j]=0.0;
  }

  // j is column
  for(j=0;j<size-1;j++){
    output[j]+=diag_R[j]*b[j];
    istart=xlnz[j];
    istop=xlnz[j+1]-1;
    i=xnzsub[j];
    for(ii=istart;ii<=istop;ii++){
      row=nzsub[i-1]-1;
      output[row]+=lnz_R[ii-1]*b[j];
      output[j]+=lnz_R[ii-1]*b[row];
      i++;
    }
  }
  output[size-1]+=diag_R[size-1]*b[size-1];
}











